# Repository path:    $HeadURL: https://90.145.156.222/DMS-Houtribdijk/code/nc-to-thredds/source/ProcessData.py $
# Last committed:     $Revision: 587939 $; $Author: vvuik $; $Date: 2020-03-27 10:01:46 +0100 (vr, 27 mrt 2020) $
# -*- coding: utf-8 -*-
"""
Created on Mon Sep  9 15:48:42 2019

@author: hove
"""

from datetime import datetime, timedelta
import os
import numpy as np
import pandas as pd

from golf_parameters import golf_parameters
from golf_MEMpuv import golf_MEMpuv

from writeNetCDF import write_location_data, write_global_data, write_standard_data, write_standard_data_levels, write_tailored_data
from readAquo import read_aquo, read_aquo_levels
from importData import import_data
from getData import get_parameters, get_parameters_levels, get_parameters_tailored, get_location_list, get_meta_data, get_coordinates
    
#%%
    
def fews_params(ncpar):
    
    return ncpar.replace('_', '.')

#%%

# STB has tailored data

def run_STB(instrument, settings, ndays, specificlocs=None):
    
    startTime = datetime.now().replace(microsecond=0,second=0,minute=0,hour=0) - timedelta(days=ndays)
    endTime = startTime + timedelta(days=1)
    
    year = startTime.year
    month = startTime.month
    day = startTime.day
    
    aquo = read_aquo()
    
    parametersRO, parametersRV, parametersNC = get_parameters(instrument)
    parameters_tailored = get_parameters_tailored(instrument)
    
    if specificlocs is None:
        location_list = get_location_list(instrument, startTime)
    else:
        location_list = specificlocs
    
    for location_id in location_list:
        
        print("Processing {}".format(location_id))
        
        # Import data STB
        timeseries, location_name = import_data(instrument, location_id, parametersRO, startTime, endTime, settings)
        
        # No timeseries, then continue with next location
        if timeseries is None:
            continue
        
        # Processing standard data specific for STB
        df = timeseries.pivot(columns='parameterId', values='value')
        df = df.interpolate(limit=2)
        # Delete endTime
        try:
            df.drop(pd.Timestamp(endTime, tz='Etc/GMT'), inplace=True)
        except:
            pass
        # Continue if no data in df
        if df.isnull().all().all() == True:
            print("At start time {} and location {}: Data in FEWS is NaN".format(startTime, location_id))
            continue
        
        # Location data
        X, Y, Z, bedlevel, tZ, SN, drukoffset = get_meta_data(settings, startTime, location_id, instrument)
        lon, lat = get_coordinates(location_id, X, Y)
        
        # convert EtroLevel into water level
        df['H.w.0'] = df['EtroLevel'] + Z
        df.drop(['EtroLevel'], axis=1, inplace=True)
        
        # Tailored data
        li, spec, fx = golf_parameters(df, startTime, endTime, instrument)    
        frequency = np.round(np.linspace(start = 0, stop = 2, num = 201), 2) # to avoid floating errors   
        # determine the limit of the interpolation region. 
        # if 0.494949, we want to extrapolate to get a value for freq = 0.5, but not for freq > 0.5
        freq_max = round(fx[-1],1)
        spectra = pd.DataFrame(index=spec.index, columns=frequency)
        for index, row in spec.iterrows():
            vy = row.vy.astype('float')
            spectra.loc[index, :] = [np.interp(freq, fx, vy) if freq<=freq_max else np.nan for freq in spectra.columns]
        
        # Write netcdf file standard data
        nc_root_dir = os.path.join(settings['ncdir_standard'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}000000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        nc = write_standard_data(nc, df, parametersNC, aquo)
        
        nc.summary = 'data measured by {} sensor at platform {}'.format(instrument, location_id)
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset,scriptdir)
        
        # Add data specific for STB
        nc.measurement_frequency = '4 Hz'
        nc.close()
        
        # Write netcdf file tailored data
        nc_root_dir = os.path.join(settings['ncdir_tailored'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}000000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        nc = write_tailored_data(nc, spectra.values, li, frequency, parameters_tailored, aquo)
        
        nc.summary = 'wave parameters from data measured by {} sensor at platform {}'.format(instrument, location_id)
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset,scriptdir)
        
        # Add data specific for STB
        nc.measurement_frequency = '4 Hz'
        nc.burst_length = '15 min'
        nc.max_overlap_Welch = '0.5'
        nc.filter = 'Hann'
        nc.close()

#%%

def run_MPP(instrument, settings, ndays, specificlocs=None):
    
    startTime = datetime.now().replace(microsecond=0,second=0,minute=0,hour=0) - timedelta(days=ndays)
    endTime = startTime + timedelta(days=1)
    
    year = startTime.year
    month = startTime.month
    day = startTime.day
    
    aquo = read_aquo()
    
    parametersRO, parametersRV, parametersNC = get_parameters(instrument)
    
    if specificlocs is None:
        location_list = get_location_list(instrument, startTime)
    else:
        location_list = specificlocs
    
    for location_id in location_list:
        
        print("Processing {}".format(location_id))
        
        if startTime < datetime(2019,8,5,0,0,0) and location_id=='BUOY1':
            print("Skipping location BUOY1")
            continue
        
        # Import data MPP
        timeseries, location_name = import_data(instrument, location_id, parametersRO, startTime, endTime, settings)
        
        # No timeseries, then continue with next location
        if timeseries is None:
            continue
        
        # Processing standard data specific for MPP
        df = timeseries.pivot(columns='parameterId', values='value')
        # Delete endTime
        try:
            df.drop(pd.Timestamp(endTime, tz='Etc/GMT'), inplace=True)
        except:
            pass
        # Continue if no data in df
        if df.isnull().all().all() == True:
            print("At start time {} and location {}: Data in FEWS is NaN".format(startTime, location_id))
            continue
        
        # Fill column for oxygen with NaN if not available in FEWS
        if 'O2.0.mgl' not in df.columns:
            df['O2.0.mgl'] = np.NaN
        if 'O2.0.perc' not in df.columns:
            df['O2.0.perc'] = np.NaN
        
        # Location data
        X, Y, Z, bedlevel, tZ, SN, drukoffset = get_meta_data(settings, startTime, location_id, instrument)
        lon, lat = get_coordinates(location_id, X, Y)
        
        # Write netcdf file
        nc_root_dir = os.path.join(settings['ncdir_standard'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}000000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        nc = write_standard_data(nc, df, parametersNC, aquo)
        
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset,scriptdir)
        nc.summary = 'data measured by {} sensor at platform {}'.format(instrument, location_id)
        
        # Add data specific for MPP
        nc.measurement_frequency = '1 Hz'
        nc.time_per_ensemble = '10 min'
        nc.close()

#%%

def run_ALT(instrument, settings, ndays, specificlocs=None):
    
    startTime = datetime.now().replace(microsecond=0,second=0,minute=0,hour=0) - timedelta(days=ndays)
    endTime = startTime + timedelta(days=1)
    
    year = startTime.year
    month = startTime.month
    day = startTime.day
    
    aquo = read_aquo()
    
    parametersRO, parametersRV, parametersNC = get_parameters(instrument)
    
    if specificlocs is None:
        location_list = get_location_list(instrument, startTime)
    else:
        location_list = specificlocs
    
    for location_id in location_list:
        
        print("Processing {}".format(location_id))
        
        # Import data ALT
        timeseries, location_name = import_data(instrument, location_id, parametersRO, startTime, endTime, settings)
        
        # No timeseries, then continue with next location
        if timeseries is None:
            continue
        
        # Processing standard data specific for ALT
        df = timeseries.pivot(columns='parameterId', values='value')
        # Delete endTime
        try:
            df.drop(pd.Timestamp(endTime, tz='Etc/GMT'), inplace=True)
        except:
            pass
        # Continue if no data in df
        if df.isnull().all().all() == True:
            print("At start time {} and location {}: Data in FEWS is NaN".format(startTime, location_id))
            continue
        
        # Location data
        X, Y, Z, bedlevel, tZ, SN, drukoffset = get_meta_data(settings, startTime, location_id, instrument)
        lon, lat = get_coordinates(location_id, X, Y)
        
        # Convert ALTdist into bed level
        df['H.b.0'] = Z - df['ALTdist']
        
        # Write netcdf file
        nc_root_dir = os.path.join(settings['ncdir_standard'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}000000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        nc = write_standard_data(nc, df, parametersNC, aquo)
        
        nc.summary = 'data measured by {} sensor at platform {}'.format(instrument, location_id)
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset,scriptdir)
        
        # Add data specific for ALT
        nc.measurement_frequency = '500000'
        nc.close()
   
#%%

# ADC has levels 

def run_ADC(instrument, settings, ndays, specificlocs=None):
    
    startTime = datetime.now().replace(microsecond=0,second=0,minute=0,hour=0) - timedelta(days=ndays)
    endTime = startTime + timedelta(days=1)
    
    year = startTime.year
    month = startTime.month
    day = startTime.day
    
    blank_size = 0.25 # blank: WF in configuratie
    nr = 24 # number of levels: WN in configuratie
    level_size = 0.25 # cell size: WS in configuratie
    
    aquo = read_aquo()
    aquo_levels = read_aquo_levels()
    
    parametersRO, parametersRV, parametersNC = get_parameters(instrument)
    parameters_levels, levels_dict = get_parameters_levels(instrument, nr)
    parameters_total = parametersRO + parameters_levels
    
    if specificlocs is None:
        location_list = get_location_list(instrument, startTime)
    else:
        location_list = specificlocs
    
    for location_id in location_list:
        
        print("Processing {}".format(location_id))
        
        # Import data ADC
        timeseries, location_name = import_data(instrument, location_id, parameters_total, startTime, endTime, settings)
        
        # No timeseries, then continue with next location
        if timeseries is None:
            continue
        
        # Processing standard data specific for ADC
        df = timeseries.pivot(columns='parameterId', values='value')
        # Delete endTime
        try:
            df.drop(pd.Timestamp(endTime, tz='Etc/GMT'), inplace=True)
        except:
            pass
        # Continue if no data in df
        if df.isnull().all().all() == True:
            print("At start time {} and location {}: Data in FEWS is NaN".format(startTime, location_id))
            continue
        
        # Location data
        X, Y, Z, bedlevel, tZ, SN, drukoffset = get_meta_data(settings, startTime, location_id, instrument)
        lon, lat = get_coordinates(location_id, X, Y)
        
        # Remove layers above water level 
        level_0 = Z + blank_size # assume Z of level 0 is instrument height
        levels = np.linspace(start=level_0, stop=level_0+level_size*nr, num=nr, endpoint=False) # instrument is looking up
        # select levels between -1.5 m+NAP and 0.5 m+NAP
        levels_sel = [index for index, value in enumerate(levels) if (value < 0.5) & (value > -1.5)]
        n_min, n_max = np.min(levels_sel), np.max(levels_sel)

        df['E.max'] = df.loc[:, 'E1.0.n{}'.format(str(n_min).zfill(2)):'E1.0.n{}'.format(str(n_max))].idxmax(axis=1, skipna=True)
        df['Boundary'] = [int(df.at[i, 'E.max'][-2:])-2 for i in df.index]
        
        fews_parameters_levels = [fews_params(w) for w in parameters_levels]

        for var in fews_parameters_levels:
            df[var] = np.where(int(var[-2:]) > df['Boundary'], np.NaN, df[var])

        df.drop(['E.max', 'Boundary'], axis=1, inplace=True)
        
        # Write netcdf file
        nc_root_dir = os.path.join(settings['ncdir_standard'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}000000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        nc = write_standard_data(nc, df, parametersNC, aquo)
        nc = write_standard_data_levels(nc, df, levels_dict, aquo_levels, levels)
        
        nc.summary = 'data measured by {} sensor at platform {}'.format(instrument, location_id)
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset, scriptdir)
        
        # Add data specific for ADC
        nc.blank_distance = '25 cm'
        nc.number_of_layers = '24'
        nc.pings_per_ensemble = '500'
        nc.layer_size = '25 cm'
        nc.ambiquity_velocity = '170 cm/s'
        nc.time_per_ensemble = '10 min'
        nc.time_between_pings = '1.09 s'
        nc.close()

#%%

# ADV has hours
# ADV has corrected pressure and waterlevels
# ADV has tailored data

def run_ADV(instrument, settings, ndays, nhours, specificlocs=None):
    
    startTime = datetime.now().replace(microsecond=0,second=0,minute=0)-timedelta(days=ndays, hours=nhours)
    endTime = startTime + timedelta(hours=1) 

    year = startTime.year
    month = startTime.month
    day = startTime.day
    hour = startTime.hour
    
    aquo = read_aquo()
    
    parametersRO, parametersRV, parametersNC = get_parameters(instrument)
    parameters_tailored = get_parameters_tailored(instrument)
    
    if specificlocs is None:
        location_list = get_location_list(instrument, startTime)
    else:
        location_list = specificlocs
    
    for location_id in location_list:
        
        print("Processing {}".format(location_id))
        
        # Import data ADV
        timeseries, location_name = import_data(instrument, location_id, parametersRO, startTime, endTime, settings)
        timeseries_val, location_name = import_data(instrument, location_id, parametersRV, startTime, endTime, settings, validated=True)
        
        # No timeseries, then continue with next location
        if timeseries is None:
            continue
        
        # If both timeseries are present, then concat the two
        if (timeseries is not None) & (timeseries_val is not None):
            timeseries_total = pd.concat([timeseries, timeseries_val], sort=True)
        else:
            timeseries_total = timeseries.copy()
        
        # Processing standard data specific for ADV
        df = timeseries_total.pivot(columns='parameterId', values='value')
        # Delete endTime
        try:
            df.drop(pd.Timestamp(endTime, tz='Etc/GMT'), inplace=True)
        except:
            pass
        # Continue if no data in df
        if df.isnull().all().all() == True:
            print("At start time {} and location {}: Data in FEWS is NaN".format(startTime, location_id))
            continue
        
        # Location data, Z is height sampling volume Vector ADV in metadatasheet
        # pressure sensor 0.368 m higher than sampling volume
        X, Y, Z, bedlevel, tZ, SN, drukoffset = get_meta_data(settings, startTime, location_id, instrument)
        dzpres = 0.368 # afstand pressure sensor boven meetvolume ADV
        Zdruksensor = Z+dzpres # hoogte druksensor (m+NAP)
        
        lon, lat = get_coordinates(location_id, X, Y)
        
        # Fill column for pressure and waterlevel with NaN if not available in FEWS
        if 'D.0.water' not in df.columns: # pressure
            df['D.0.water'] = np.NaN
        if 'H.w.0' not in df.columns: # water level
            df['H.w.0'] = np.NaN
        
        ######## Tailored data ########
        
        if location_id == 'BUOY1':
            # No tailored data for BUOY1
            print('skip tailored data for Frame Marker Wadden (BUOY1)')
            waves = False
        else:        
            # add additional columns for validation        
            df['V.total'] = ((df['S.0.east']**2 + df['S.0.north']**2 + df['S.0.up']**2).values)**0.5
            df['V.previous'] = df['V.total'].shift(1)
            df['D.previous'] = df['D.0.water'].shift(1)
            
            Fs = 4 # sampling frequency of sensor 
            bLen = 15 # length of sample in minutes
            maxE = 0.01 # maximum fraction of empty values in bin for computation of waves
            bNum = bLen*60*Fs # number of samples in bin
            eMax = np.ceil(maxE*bNum) # maximum number of empty values in bin
            
            # bins for validation as well as waves
            date_range = pd.date_range(start = startTime, end = endTime, freq = str(bLen)+'min')
            
            li_total = []
            vy_total = []
            fx_check = []
            S_total = np.zeros((len(date_range)-1, 64, 64))
    
            for i in range(len(date_range)-1):
                
                # define a bin range and reindex 
                bin_range = pd.date_range(start = date_range[i], end = date_range[i+1], freq = '250ms', closed = 'left', tz = 'Etc/GMT')
                df_temp = df.reindex(bin_range)
                
                # check for outliers in pressure
                if df_temp['D.0.water'].isnull().all() != True:
                    
                    mu = np.nanmean(df_temp['D.0.water'].astype(float))
                    sigma = np.nanstd(df_temp['D.0.water'].astype(float))
                    limit = mu + 4*sigma
                    df_temp['D.0.water'] = [i if i <= limit else np.nan for i in df_temp['D.0.water'].values] 
                    df_temp['D.0.water'] = [df_temp.at[i, 'D.0.water'] if abs(df_temp.at[i, 'D.0.water'] - df_temp.at[i, 'D.previous']) < (4 * sigma) else np.nan for i in df_temp.index]
                
                # check for outliers in velocity
                if df_temp['V.total'].isnull().all() != True:
                            
                    mu = np.nanmean(df_temp['V.total'].astype(float))
                    sigma = np.nanstd(df_temp['V.total'].astype(float))
                    limit = mu + 4*sigma
                    df_temp['V.total'] = [i if i <= limit else np.nan for i in df_temp['V.total'].values] 
                    df_temp['V.total'] = [df_temp.at[i, 'V.total'] if abs(df_temp.at[i, 'V.total'] - df_temp.at[i, 'V.previous']) < (4 * sigma) else np.nan for i in df_temp.index]
                        
                    idx_isnull = df_temp['V.total'].isnull()
                    df_temp.loc[idx_isnull, ['S.0.east', 'S.0.north', 'S.0.up']] = np.nan
                
                # compute amount of nan values
                eNan_D = df_temp['D.0.water'].isnull().sum() 
                eNan_U = df_temp['V.total'].isnull().sum()
        
                if eNan_D < eMax and eNan_U < eMax:
            
                    # interpolate missing values
                    p = df_temp['D.0.water'].interpolate(method = 'time', limit_direction = 'both').ffill().bfill().values
                    u = df_temp['S.0.east'].interpolate(method = 'time', limit_direction = 'both').ffill().bfill().values
                    v = df_temp['S.0.north'].interpolate(method = 'time', limit_direction = 'both').ffill().bfill().values
                    
                    # compute wave data. Error made: assume that u and v are measured at the position of the pressure sensor
                    li, vy, fx, S = golf_MEMpuv(_p=p, _u=u, _v=v, sensorlevel=Zdruksensor, bedlevel=bedlevel)
                    
                    li_total.append(li) # includes hmean, Hm0, Tp, Tm01, Tm02, Tmm10, theta0, dspr (value per bin)
                    vy_total.append(vy) # Czz: wave variance spectrum
                    fx_check.append(fx) # frequency (should be the same for every bin, right?)
                    S_total[i][:][:] = S # 2D spectrum
                    
                else:
                    # append arrays with nan 
                    li_total.append(8*[np.NaN])
                    vy_total.append(64*[np.NaN])
                    S_total[i][:][:] = np.ones((64,64))*np.NaN
                
            # array with vy for all bins
            vy_array = np.vstack(vy_total)
            # dataframe with wave parameters for all bins
            li = pd.DataFrame(data = li_total, columns = ['hmean', 'Hm0', 'Tp', 'Tm01', 'Tm02', 'Tmm10', 'theta0', 'dspr'], index = date_range[:-1])
    
            # theta = wave direction
            # this parameter is defined in golf_MEMpuv
            ntheta  = 64
            thetmax = np.pi
            dtheta  = 2*thetmax/ntheta
            theta   = np.arange(start=-thetmax,stop=thetmax,step=dtheta)
                
            # frequency        
            if len(fx_check) == 0:
                # no wave parameters were computed for any of the bins (too many NaN)
                print("{} and {}: no wave parameters".format(startTime, location_id))
                waves = False
            elif all((fx_check[i] == fx_check[0]).all() for i in range(len(fx_check))) == True:
                frequency = fx_check[0]
                waves = True
            else:
                print("error: the computed frequency is not constant")
                waves = False
        
        # Write netcdf file
        nc_root_dir = os.path.join(settings['ncdir_standard'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}{}0000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2), str(hour).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        nc = write_standard_data(nc, df, parametersNC, aquo)
        
        nc.summary = 'data measured by {} sensor at platform {}'.format(instrument, location_id)
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset, scriptdir)
        
        # Add data specific for ADV
        nc.measurement_frequency = '4 Hz'
        nc.close()
        
        # Write netcdf file tailored data
        if waves == True:
            nc_root_dir = os.path.join(settings['ncdir_tailored'], instrument, location_id)
            ncfile = '{}_{}_{}{}{}{}0000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2), str(hour).zfill(2))
            filename = os.path.join(nc_root_dir, ncfile)
              
            nc = write_location_data(filename, X, Y, lon, lat)
            nc = write_tailored_data(nc, vy_array, li, frequency, parameters_tailored, aquo, theta, S_total)
            
            nc.summary = 'wave parameters from data measured by {} sensor at platform {}'.format(instrument, location_id)
            scriptdir = settings['scripts_directory']
            nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset, scriptdir)
            
            # Add data specific for ADV
            nc.measurement_frequency = '4 Hz'
            nc.burst_length = '15 min'
            nc.cross_power_spectral_density = 'Welch'
            nc.filter = 'Hann'
            nc.linear_dispersion_relation = 'Klopman'
            nc.close()

#%%

# AQD has levels 
# AQD has corrected pressure and waterlevels
# AQD has tailored data

# Instrument is looking down. This is different from ADC 
# Numbering of layers is different from ADC 

def run_AQD(instrument, settings, ndays, nhours, specificlocs=None):
    
    startTime = datetime.now().replace(microsecond=0,second=0,minute=0)-timedelta(days=ndays, hours=nhours)
    endTime = startTime + timedelta(hours=1) 

    year = startTime.year
    month = startTime.month
    day = startTime.day
    hour = startTime.hour
    
    blank_size = 0.1 # m
    nr = 15 
    level_size = 0.03 # m
    
    aquo = read_aquo()
    aquo_levels = read_aquo_levels()
    
    parametersRO, parametersRV, parametersNC = get_parameters(instrument)
    parameters_levels, levels_dict = get_parameters_levels(instrument, nr)
    parameters_total = parametersRO + parameters_levels
    parameters_tailored = get_parameters_tailored(instrument)
    
    if specificlocs is None:
        location_list = get_location_list(instrument, startTime)
    else:
        location_list = specificlocs
    
    for location_id in location_list:
        
        print("Processing {}".format(location_id))
        
        # Import data AQD
        timeseries, location_name = import_data(instrument, location_id, parameters_total, startTime, endTime, settings)
        timeseries_val, location_name = import_data(instrument, location_id, parametersRV, startTime, endTime, settings, validated=True)
        
        # No timeseries, then continue with next location
        if timeseries is None:
            print("No data in FEWS")
            continue
        
        # If both timeseries are present, then concat the two
        if (timeseries is not None) & (timeseries_val is not None):
            timeseries_total = pd.concat([timeseries, timeseries_val], sort=True)
        else:
            timeseries_total = timeseries.copy()
        
        # Processing standard data specific for AQD
        df = timeseries_total.pivot(columns='parameterId', values='value')
        # Delete endTime
        try:
            df.drop(pd.Timestamp(endTime, tz='Etc/GMT'), inplace=True)
        except:
            pass
        # Continue if no data in df
        if df.isnull().all().all() == True:
            print("At start time {} and location {}: Data in FEWS is NaN".format(startTime, location_id))
            continue
        
        # Location data
        X, Y, Z, bedlevel, tZ, SN, drukoffset = get_meta_data(settings, startTime, location_id, instrument)
        lon, lat = get_coordinates(location_id, X, Y)
        
        # Layers
        level_0 = Z - blank_size # assume Z of level 0 is instrument height
        levels = np.linspace(start=level_0-level_size*nr, stop=level_0, num=nr, endpoint=False) # instrument is looking down and first in list is bottom layer (meta data netcdf)
        
        # Write netcdf file
        nc_root_dir = os.path.join(settings['ncdir_standard'], instrument, location_id)
        ncfile = '{}_{}_{}{}{}{}0000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2), str(hour).zfill(2))
        filename = os.path.join(nc_root_dir, ncfile)
        
        nc = write_location_data(filename, X, Y, lon, lat)
        
        if 'D.0.water' in df.columns and 'H.w.0' in df.columns:
            nc = write_standard_data(nc, df, parametersNC, aquo)
        else:
            nc = write_standard_data(nc, df, parametersRO, aquo)
            print('warning: no water levels in FEWS, generating netcdf without pressure and water level')
        nc = write_standard_data_levels(nc, df, levels_dict, aquo_levels, levels)
        
        nc.summary = 'data measured by {} sensor at platform {}'.format(instrument, location_id)
        scriptdir = settings['scripts_directory']
        nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset, scriptdir)
        
        # Add data specific for AQD
        nc.measurement_frequency = '4 Hz (systeemdata: 1 Hz)'
        nc.blank_distance = '10 cm'
        nc.number_of_layers = '15'
        nc.pulse_distance = '60 cm'
        nc.layer_size = '3 cm'
        nc.close()
        
        ######### Tailored data ################
        
        if 'H.w.0' not in df.columns:
            print("Warning: No H.w.0 in FEWS, no tailored data")
            continue
        else:
            df = timeseries_val.pivot(columns='parameterId', values='value')
            li, spec, fx = golf_parameters(df, startTime, endTime, instrument)
            if fx is None:
                print('no wave spectra could be computed for location {}'.format(location_id))
                continue
            frequency = np.round(np.linspace(start = 0, stop = 2, num = 201), 2) # to avoid floating errors   
            # determine the limit of the interpolation region. 
            # if 0.494949, we want to extrapolate to get a value for freq = 0.5, but not for freq > 0.5
            freq_max = round(fx[-1],1)
            spectra = pd.DataFrame(index=spec.index, columns=frequency)
            for index, row in spec.iterrows():
                vy = row.vy.astype('float')
                spectra.loc[index, :] = [np.interp(freq, fx, vy) if freq<=freq_max else np.nan for freq in spectra.columns]
            
            # Write netcdf file tailored data
            nc_root_dir = os.path.join(settings['ncdir_tailored'], instrument, location_id)
            ncfile = '{}_{}_{}{}{}{}0000.nc'.format(instrument, location_id, year, str(month).zfill(2), str(day).zfill(2), str(hour).zfill(2))
            filename = os.path.join(nc_root_dir, ncfile)
            
            nc = write_location_data(filename, X, Y, lon, lat)
            nc = write_tailored_data(nc, spectra, li, frequency, parameters_tailored, aquo)
            
            nc.summary = 'wave parameters from data measured by {} sensor at platform {}'.format(instrument, location_id)
            scriptdir = settings['scripts_directory']
            nc = write_global_data(nc, instrument, location_id, location_name, startTime, endTime, Z, bedlevel, tZ, SN, drukoffset, scriptdir)
            
            # Add data specific for AQD
            nc.measurement_frequency = '4 Hz'
            nc.burst_length = '14 min'
            nc.max_overlap_Welch = '0.5'
            nc.filter = 'Hann'
            nc.close()
            
